@@ -75,6 +75,13 @@ def getChildrenByName(dom, namespace, name):
                             n.localName == name,
                   dom.childNodes)
 
+def getChildrenByNameAndAttribute(dom, namespace, name, attribute, value):
+    return filter(lambda n: n.nodeType == n.ELEMENT_NODE and \
+                            n.namespaceURI == namespace and \
+                            n.localName == name and \
+                            n.getAttribute(attribute) == value,
+                  dom.childNodes)
+
 def getOnlyChildByName(dom, namespace, name):
     kids = getChildrenByName(dom, namespace, name)
 
@@ -88,6 +95,19 @@ def getOnlyChildByName(dom, namespace, name):
 
     return kids[0]
 
+def getAnnotationByName(dom, name):
+    kids = getChildrenByNameAndAttribute(dom, None, 'annotation', 'name', name)
+
+    if len(kids) == 0:
+        return None
+
+    if len(kids) > 1:
+        raise WrongNumberOfChildren(
+            '<%s> node should have at most one %s annotation' %
+            (dom.tagName, name))
+
+    return kids[0].getAttribute('value')
+
 def getNamespace(n):
     if n.namespaceURI is not None:
         return n.namespaceURI
@@ -142,6 +162,10 @@ class Base(object):
         self.docstring = getOnlyChildByName(dom, XMLNS_TP, 'docstring')
         self.added = getOnlyChildByName(dom, XMLNS_TP, 'added')
         self.deprecated = getOnlyChildByName(dom, XMLNS_TP, 'deprecated')
+        if self.deprecated is None:
+            self.is_deprecated = (getAnnotationByName(dom, 'org.freedesktop.DBus.Deprecated') == 'true')
+        else:
+            self.is_deprecated = True
 
         self.changed = getChildrenByName(dom, XMLNS_TP, 'changed')
 
@@ -211,8 +235,14 @@ class Base(object):
                                           "Added in %s.")
 
     def get_deprecated(self):
-        return self._get_generic_with_ver(self.deprecated, 'deprecated',
-                                          "Deprecated since %s.")
+        if self.deprecated is None:
+            if self.is_deprecated:
+                return '<div class="annotation deprecated no-version">Deprecated.</div>'
+            else:
+                return ''
+        else:
+            return self._get_generic_with_ver(self.deprecated, 'deprecated',
+                                              "Deprecated since %s.")
 
     def get_changed(self):
         return '\n'.join(map(lambda n:
@@ -400,6 +430,15 @@ WARNING: Key '%s' not known in namespace '%s'
     def __repr__(self):
         return '%s(%s)' % (self.__class__.__name__, self.name)
 
+    def get_index_entries(self):
+        context = self.parent.get_index_context()
+        return set([
+            '%s (%s in %s)' % (self.short_name, self.get_type_name(), context),
+            '%s %s' % (self.get_type_name(), self.name)])
+
+    def get_index_context(self):
+        return self.short_name
+
 class DBusConstruct(Base):
     """Base class for signals, methods and properties."""
 
@@ -471,6 +510,8 @@ class Method(DBusConstruct):
         self.possible_errors = build_list(self, PossibleError, None,
                         dom.getElementsByTagNameNS(XMLNS_TP, 'error'))
 
+        self.no_reply = (getAnnotationByName(dom, 'org.freedesktop.DBus.Method.NoReply') == 'true')
+
     def get_in_args(self):
         return ', '.join(map(lambda a: a.spec_name(), self.in_args))
 
@@ -543,11 +584,16 @@ class Typed(Base):
         return '%s(%s:%s)' % (self.__class__.__name__, self.name, self.dbus_type)
 
 class Property(DBusConstruct, Typed):
-    ACCESS_READ     = 1
-    ACCESS_WRITE    = 2
+    ACCESS_READ = 1
+    ACCESS_WRITE = 2
 
     ACCESS_READWRITE = ACCESS_READ | ACCESS_WRITE
 
+    EMITS_CHANGED_UNKNOWN = 0
+    EMITS_CHANGED_NONE = 1
+    EMITS_CHANGED_UPDATES = 2
+    EMITS_CHANGED_INVALIDATES = 3
+
     def __init__(self, parent, namespace, dom):
         super(Property, self).__init__(parent, namespace, dom)
 
@@ -572,6 +618,21 @@ class Property(DBusConstruct, Typed):
         self.requestable = requestable != ''
         self.sometimes_requestable = requestable == 'sometimes'
 
+        # According to the D-Bus specification, EmitsChangedSignal defaults
+        # to true, but - realistically - this cannot be assumed for old specs.
+        # As a result, we treat the absence of the annotation as "unknown".
+        emits_changed = getAnnotationByName(dom, 'org.freedesktop.DBus.Property.EmitsChangedSignal')
+        if emits_changed is None:
+            emits_changed = getAnnotationByName(dom.parentNode, 'org.freedesktop.DBus.Property.EmitsChangedSignal')
+        if emits_changed == 'true':
+            self.emits_changed = self.EMITS_CHANGED_UPDATES;
+        elif emits_changed == 'invalidates':
+            self.emits_changed = self.EMITS_CHANGED_INVALIDATES;
+        elif emits_changed == 'false':
+            self.emits_changed = self.EMITS_CHANGED_NONE;
+        else:
+            self.emits_changed = self.EMITS_CHANGED_UNKNOWN;
+
     def get_access(self):
         if self.access & self.ACCESS_READ and self.access & self.ACCESS_WRITE:
             return 'Read/Write'
@@ -764,6 +825,7 @@ WARNING: Interface not known: '%s'
 
     def check_for_odd_children(self, dom):
         expected = [
+            (None, 'annotation'),
             (None, 'method'),
             (None, 'property'),
             (None, 'signal'),
@@ -1102,7 +1164,7 @@ class SectionBase(object):
     """
 
     def __init__(self, dom, spec_namespace):
-
+        self.spec_namespace = spec_namespace
         self.items = []
 
         def recurse(nodes):
@@ -1123,6 +1185,9 @@ class SectionBase(object):
 
         recurse(dom.childNodes)
 
+    def get_index_context(self):
+        return self.spec_namespace
+
 class Section(Base, SectionBase):
     def __init__(self, parent, namespace, dom, spec_namespace):
         Base.__init__(self, parent, namespace, dom)
@@ -1139,6 +1204,7 @@ class Spec(SectionBase):
     def __init__(self, dom, spec_namespace, allow_externals=False):
         self.document = dom
         self.spec_namespace = spec_namespace
+        self.short_name = spec_namespace
         self.allow_externals = allow_externals
 
         # build a dictionary of errors in this spec
